
/* Generated data (by glib-mkenums) */


#ifndef POPPLER_ENUMS_H
#define POPPLER_ENUMS_H

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "poppler-action.h" */
GType poppler_action_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_ACTION_TYPE (poppler_action_type_get_type ())
GType poppler_dest_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_DEST_TYPE (poppler_dest_type_get_type ())

/* enumerations from "poppler-document.h" */
GType poppler_page_layout_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_PAGE_LAYOUT (poppler_page_layout_get_type ())
GType poppler_page_mode_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_PAGE_MODE (poppler_page_mode_get_type ())
GType poppler_font_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_FONT_TYPE (poppler_font_type_get_type ())
GType poppler_viewer_preferences_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_VIEWER_PREFERENCES (poppler_viewer_preferences_get_type ())
GType poppler_permissions_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_PERMISSIONS (poppler_permissions_get_type ())

/* enumerations from "poppler-form-field.h" */
GType poppler_form_field_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_FORM_FIELD_TYPE (poppler_form_field_type_get_type ())
GType poppler_form_button_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_FORM_BUTTON_TYPE (poppler_form_button_type_get_type ())
GType poppler_form_text_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_FORM_TEXT_TYPE (poppler_form_text_type_get_type ())
GType poppler_form_choice_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_FORM_CHOICE_TYPE (poppler_form_choice_type_get_type ())

/* enumerations from "poppler-annot.h" */
GType poppler_annot_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_ANNOT_TYPE (poppler_annot_type_get_type ())
GType poppler_annot_flag_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_ANNOT_FLAG (poppler_annot_flag_get_type ())
GType poppler_annot_markup_reply_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_ANNOT_MARKUP_REPLY_TYPE (poppler_annot_markup_reply_type_get_type ())
GType poppler_annot_external_data_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_ANNOT_EXTERNAL_DATA_TYPE (poppler_annot_external_data_type_get_type ())
GType poppler_annot_text_state_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_ANNOT_TEXT_STATE (poppler_annot_text_state_get_type ())
GType poppler_annot_free_text_quadding_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_ANNOT_FREE_TEXT_QUADDING (poppler_annot_free_text_quadding_get_type ())

/* enumerations from "poppler.h" */
GType poppler_error_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_ERROR (poppler_error_get_type ())
GType poppler_orientation_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_ORIENTATION (poppler_orientation_get_type ())
GType poppler_page_transition_type_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_PAGE_TRANSITION_TYPE (poppler_page_transition_type_get_type ())
GType poppler_page_transition_alignment_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_PAGE_TRANSITION_ALIGNMENT (poppler_page_transition_alignment_get_type ())
GType poppler_page_transition_direction_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_PAGE_TRANSITION_DIRECTION (poppler_page_transition_direction_get_type ())
GType poppler_selection_style_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_SELECTION_STYLE (poppler_selection_style_get_type ())
GType poppler_backend_get_type (void) G_GNUC_CONST;
#define POPPLER_TYPE_BACKEND (poppler_backend_get_type ())
G_END_DECLS

#endif /* !POPPLER_ENUMS_H */

/* Generated data ends here */

